import os
import tempfile
from io import StringIO
import sys
from appdirs import AppDirs
from unittest.mock import patch
import pytest
import zml
from zml.node import Egg, Node, TreeNode, Model
from zml.context import RenderingContext
from zml.document import Document, DocumentRenderingContext, DocumentLookup
from zml.util import minimise
from zml.exceptions import (DocumentNotDefinedException,
                            FileNotLoadedException,
                            IndentationException,
                            VariableNotDefinedException,
                            TranslationNotDefinedException)


def run_test(code, target, global_context={}):
    # render code
    out = zml.render(code, global_context=global_context)
    assert minimise(out) == minimise(target)
    # render file
    temp = tempfile.NamedTemporaryFile(suffix='.zml', mode='w+', encoding='utf-8', delete=False)
    temp.write(code)
    temp.close()
    out = zml.render(temp.name, global_context=global_context)
    os.unlink(temp.name)
    assert minimise(out) == minimise(target)


class TestZml(object):

    def setup(self):
        base_path = os.path.dirname(os.path.abspath(__file__))
        zml.set_base_path(base_path)

    def test_module_render_with_code_param_returns_string(self):
        code = "'ZML'"
        result = zml.render(code)
        assert isinstance(result, str)

    def test_document_render_with_code_param_returns_string(self):
        code = "'ZML'"
        t = Document()
        result = t.render(source=code)
        assert isinstance(result, str)

    def test_document_render_code_with_code_param_returns_string(self):
        code = "'ZML'"
        t = Document()
        result = t.render_source(source=code)
        assert isinstance(result, str)

    def test_document_render_with_document_param_returns_string(self):
        filename = 'test_plain_text.zml'
        t = Document(filename)
        result = t.render()
        assert isinstance(result, str)

    def test_render_accepts_string(self):
        code = "'ZML'"
        result = zml.render(code)
        assert isinstance(result, str)

    @patch('zml.core.Document.render')
    def xtest_render_with_filename_calls_document_render(self, mock_render):
        filename = 'test/test_plain_text.zml'
        zml.render(filename)
        assert mock_render.called is True

    def test_documentlookup_get_document_with_wrong_path_raised_exception(self):
        with pytest.raises(DocumentNotDefinedException):
            documentfile = 'test/doesnotexist.zml'
            lookup = DocumentLookup(['.'])
            lookup.get_document(documentfile)

    def test_documentlookup_get_document_returns_document(self):
        documentfile = 'test_plain_text.zml'
        lookup = DocumentLookup(['.'])
        t = lookup.get_document(documentfile)
        assert isinstance(t, Document)

    def test_render_without_document_raises_exception(self):
        with pytest.raises(DocumentNotDefinedException):
            t = Document()
            t.render()

    def test_render_with_wrong_filename_raises_exception(self):
        with pytest.raises(DocumentNotDefinedException):
            t = Document('some_path_not_exists.zml')
            t.render()

    def test_missing_file_raises_exception(self):
        with pytest.raises(DocumentNotDefinedException):
            t = Document('some_path_not_exists.zml')
            t.render()

    @patch('builtins.open')
    def test_wrong_encoded_file_raises_exception(self, mock_open):
        mock_open.side_effect = IOError
        with pytest.raises(FileNotLoadedException):
            t = Document('test_file_wrong_encoding.zml')
            t.render()

    def test_wrong_indentation_raises_exception(self):
        code = \
"""
*a
  b
    d
 c
    e
"""
        with pytest.raises(IndentationException):
            document = Document(code)
            document.source_to_tree(code)

    def test_attributes(self):
        code = \
"""
div somekey='somevalue' secondkey='secondvalue'
"""
        target = \
"""
<div somekey="somevalue" secondkey="secondvalue">
</div>
"""
        run_test(code, target)

    def test_attributes2(self):
        code = \
"""
a href='article1.html': 'more'
"""
        target = \
"""
<a href="article1.html">more</a>
"""
        run_test(code, target)

    def test_empty_attributes(self):
        code = \
"""
form:
  input type='text' disabled
"""
        target = \
"""
<form>
  <input type="text" disabled>
</form>
"""
        run_test(code, target)

    def test_all(self):
        code = \
"""
div#myid.someclass.secondclass data_one='first_value' data_two='second_value' selected
"""
        target = \
"""
<div id="myid" class="someclass secondclass" data_one="first_value" data_two="second_value" selected>
</div>
"""
        run_test(code, target)

    def test_id_classes(self):
        code = \
"""
div#myid.someclass.secondclass
"""
        target =  \
"""
<div id="myid" class="someclass secondclass">
</div>
"""
        run_test(code, target)

    def test_inline_content(self):
        code = \
"""
p: 'Some awesome text'
"""
        target = \
"""
<p>Some awesome text
</p>
"""
        run_test(code, target)

    def test_elements(self):
        code = \
"""
h1: 'One headline'
"""
        target = \
"""
<h1>One headline</h1>
"""
        run_test(code, target)

    def test_multiple(self):
        code = \
"""
p.intro: 'A teaser intro'
h1: 'One headline'
img src='article.jpg'
p: 'Some article text...'
a href='article1.html': 'more'
"""
        target = \
"""
<p class="intro">A teaser intro
</p>
<h1>One headline</h1>
<img src="article.jpg">
<p>Some article text...
</p>
<a href="article1.html">more</a>
"""
        run_test(code, target)

    def test_inline_semantics(self):
        code = \
"""
p: 'View this <strong.green data-info="cite": simple> rules to assign inline semantics to text.'
"""
        target = \
"""
<p>View this <strong class="green" data-info="cite">simple</strong> rules to assign inline semantics to text.
</p>
"""
        run_test(code, target)

    def test_text_content_nodes(self):
        code = \
"""
p:
  : 'There are some'
  strong: 'simple'
  : 'rules to assign inline semantics to text.'
  : 'This bodytext has some'
  span.highlight:
    em.red data-info='animate': 'important words'
  : 'which are emphasized.'
"""
        target = \
"""
<p>
  There are some
  <strong>simple</strong>
  rules to assign inline semantics to text.
  This bodytext has some
  <span class="highlight">
    <em class="red" data-info="animate">important words</em>
  </span>
  which are emphasized.
</p>
"""
        run_test(code, target)

    def test_nesting_structures(self):
        code = \
"""
div#sidebar:
  div.teaser:
    p.intro: 'A teaser intro'
    h1: 'One headline'
    img src='article.jpg'
    p: 'Some article text'
a href='article1.html': 'more'
"""
        target = \
"""
<div id="sidebar">
  <div class="teaser">
    <p class="intro">A teaser intro
    </p>
    <h1>One headline</h1>
    <img src="article.jpg">
    <p>Some article text
    </p>
  </div>
</div>
<a href="article1.html">more</a>
"""
        run_test(code, target)

    def test_moustache_context_item(self):
        context = {
            'title': 'Some headline in the title variable'
        }
        code = \
"""
h1: '{title}'
"""
        target = \
"""
<h1>{title}</h1>
"""
        target = target.format(title=context['title'])
        run_test(code, target, context)

    def test_moustache_context_item_with_property(self):
        context = {
            'user': {
                'firstname': 'Richard',
                'lastname': 'Langly',
                'email': 'ringo@l4ngly.org'
            }
        }
        code = \
"""
div.card:
  p: '{#user.firstname}'
  p: '{#user.lastname}'
  p: '{#user.email}'

"""
        target = \
"""
<div class="card">
  <p>{firstname}
  </p>
  <p>{lastname}
  </p>
  <p>{email}
  </p>
</div>
"""
        target = target.format(**context['user'])
        run_test(code, target, context)

    def test_inherit_document(self):
        code = \
"""
%inherit 2col

*col1_content:
  div.panel:
    h1: 'User'
"""
        target = \
"""
<html>
  <head>
    <title>zml</title>
  </head>
  <body>
    <h1>zml - zero markup language</h1>
    <div class="grid">
      <div class="m66">
        <div class="left">
          <div class="panel">
            <h1>User</h1>
          </div>
        </div>
      </div>
      <div class="m33">
        <div class="right">some sidebar stuff
        </div>
      </div>
    </div>
  </body>
</html>
"""
        run_test(code, target)

    def test_routes(self):
        code = \
"""
%import components
%inherit base

~main:
  list: '/blog/posts'
  show: '/blog/post/{id}'
  edit: '/blog/post/{id}/edit'

*content:
  ul:
    li:
      base-linkto action='list': 'List'
    li:
      base-linkto action='show' id=1: 'Details'
    li:
      base-linkto action='edit' id=1: 'Edit'
"""
        target = \
"""
<html>
  <head>
    <title>zml</title>
  </head>
  <body>
    <ul>
      <li>
        <a href="/blog/posts">List</a>
      </li>
      <li>
        <a href="/blog/post/1">Details</a>
      </li>
      <li>
        <a href="/blog/post/1/edit">Edit</a>
      </li>
    </ul>
  </body>
</html>
"""
        run_test(code, target)

    def test_components(self):
        context = {'page': {'title': 'Some title',
                            'stylesheets': ['main.css', 'content.css'],
                            'scripts': ['main.js', 'content.js']
                            },
                   'pages': [
            {'title': 'Page 1'},
            {'title': 'Page 2'},
            {'title': 'Page 3'},
        ]
        }
        code = \
"""
%import components

html:
  head:
    title: 'zml'
    %for style in #page.stylesheets:
      base-style src='{style}'
    %for script in #page.scripts:
      script src='{script}'
  body:
    base-menu items=.pages
"""
        target = \
"""
<html>
  <head>
    <title>zml</title>
    <link rel="stylesheet" type="text/css" href="main.css">
    <link rel="stylesheet" type="text/css" href="content.css">
    <script src="main.js">
    </script>
    <script src="content.js">
    </script>
  </head>
  <body>
    <div class="menu">
      <ul class="navitems">
        <li>Page 1
        </li>
        <li>Page 2
        </li>
        <li>Page 3
        </li>
      </ul>
    </div>
  </body>
</html>
"""
        run_test(code, target, context)

    def test_list_items(self):
        code = \
"""
#nodes:
  - 'first'
  - 'second'
  - 'third'

%for node in nodes:
  div: '{node}'
"""
        target = \
"""
<div>first
</div>
<div>second
</div>
<div>third
</div>
"""
        run_test(code, target)

    def test_list_items_with_dicts(self):
        code = \
"""
%import components
%inherit base

#users:
  -
    firstname: 'Richard'
    lastname: 'Langly'
    email: 'ringo@l4ngly.org'
    active: True
  -
    firstname: 'Melvin'
    lastname: 'Frohike'
    email: 'melvin@frohike1.net'
    active: True
  -
    firstname: 'John Fitzgerald'
    lastname: 'Byers'
    email: 'jfb@byers23.org'
    active: True

*content:
  %for user in #users:
    div.card:
      %if #user.active:
        p: '{#user.firstname}'
        p: '{#user.lastname}'
        p: '{#user.email}'
      %else:
        p: 'The user is not active'
"""
        target = \
"""
<html>
  <head>
    <title>zml</title>
  </head>
  <body>
    <div class="card">
      <p>Richard
      </p>
      <p>Langly
      </p>
      <p>ringo@l4ngly.org
      </p>
    </div>
    <div class="card">
      <p>Melvin
      </p>
      <p>Frohike
      </p>
      <p>melvin@frohike1.net
      </p>
    </div>
    <div class="card">
      <p>John Fitzgerald
      </p>
      <p>Byers
      </p>
      <p>jfb@byers23.org
      </p>
    </div>
  </body>
</html>
"""
        run_test(code, target)

    def test_data_sections(self):
        code = \
"""
%import components
%inherit base

#users:
  -
    firstname: 'Richard'
    lastname: 'Langly'
    email: 'ringo@l4ngly.org'
    active: True
  -
    firstname: 'Melvin'
    lastname: 'Frohike'
    email: 'melvin@frohike1.net'
    active: True
  -
    firstname: 'John Fitzgerald'
    lastname: 'Byers'
    email: 'jfb@byers23.org'
    active: True

#pages:
  -
    title: 'About'
    url: '/about'
  -
    title: 'Services'
    url: '/services'
  -
    title: 'Contact'
    url: '/contact'

#page:
  stylesheets:
    - 'files/css/base.css'
    - 'files/css/content.css'
  scripts:
    - 'files/js/jquery.js'
    - 'files/js/main.js'

*content:
  %for user in #users:
    div.card:
      %if #user.active:
        p: '{#user.firstname}'
        p: '{#user.lastname}'
        p: '{#user.email}'
      %else:
        p: 'The user is not active'
"""
        target = \
"""
<html>
  <head>
    <title>zml</title>
  </head>
  <body>
    <div class="card">
      <p>Richard
      </p>
      <p>Langly
      </p>
      <p>ringo@l4ngly.org
      </p>
    </div>
    <div class="card">
      <p>Melvin
      </p>
      <p>Frohike
      </p>
      <p>melvin@frohike1.net
      </p>
    </div>
    <div class="card">
      <p>John Fitzgerald
      </p>
      <p>Byers
      </p>
      <p>jfb@byers23.org
      </p>
    </div>
  </body>
</html>
"""
        run_test(code, target)

    def test_translations(self):
        code = \
"""
%import components
%inherit base

!en:
  labels:
    title: 'Title'
    date: 'Date'
    bodytext: 'Bodytext'
  save: 'Save'
!de:
  labels:
    title: 'Titel'
    date: 'Datum'
    bodytext: 'Haupttext'
  save: 'Speichern'

*content:
  form:
    div.formrow:
      label: !labels.title
      input type='text' name='title'
    div.formrow:
      label: !labels.bodytext
      textarea name='bodytext'
    button type='submit': !save

"""
        target = \
"""
<html>
  <head>
    <title>zml</title>
  </head>
  <body>
    <form>
      <div class="formrow">
        <label>Title</label>
        <input type="text" name="title">
      </div>
      <div class="formrow">
        <label>Bodytext</label>
        <textarea name="bodytext">
        </textarea>
      </div>
      <button type="submit">Save</button>
    </form>
  </body>
</html>
"""
        run_test(code, target)

    def test_models(self):
        code = \
"""
%import components
%inherit base

!en:
  labels:
    title: 'Title'
    date: 'Date'
    bodytext: 'Bodytext'
  buttons:
    save: 'Save'
!de:
  labels:
    title: 'Titel'
    date: 'Datum'
    bodytext: 'Haupttext'
  buttons:
    save: 'Speichern'

+post:
  .title:
    &label: !labels.title
    &type: 'str'
  .date:
    &label: !labels.date
    &type: 'datetime'
  .bodytext:
    &label: !labels.bodytext
    &type: 'str'


*content:
  base-form model=+post
"""
        target = \
"""
<html>
  <head>
    <title>zml</title>
  </head>
  <body>
    <form>
      <div class="formrow">
        <label>Title</label>
        <input type="text" name="title">
      </div>
      <div class="formrow">
        <label>Date</label>
        <input type="text" name="date">
      </div>
      <div class="formrow">
        <label>Bodytext</label>
        <input type="text" name="bodytext">
      </div>
      <button type="submit">Save</button>
    </form>
  </body>
</html>
"""
        run_test(code, target)

    def test_debug_function(self):
        capturedOutput = StringIO()
        sys.stdout = capturedOutput
        from zml.util import deb
        from zml.semantic import ELEMENT
        deb(ELEMENT.parseString("h1: 'test'"))
        sys.stdout = sys.__stdout__
        assert 'inline_content' in capturedOutput.getvalue()

    def test_rendering_context_exception(self):
        with pytest.raises(VariableNotDefinedException):
            rc = RenderingContext()
            # create empty local and global context
            rc.local_context = {}
            rc.global_context = {}
            rc.get_var('some_var_not_defined')

    def test_document_rendering_context_exception(self):
        trc = DocumentRenderingContext()
        # create empty translations
        trc.translations = {}
        with pytest.raises(TranslationNotDefinedException):
            trc.get_translation('some_translation_not_defined')
        # create single language with empty translations
        trc.translations = {'en': {}}
        with pytest.raises(TranslationNotDefinedException):
            trc.get_translation('some_translation_not_defined')

    def test_document_rendering_context_set_translation(self):
        trc = DocumentRenderingContext()
        trc.set_translation('en', 'send', 'Send')
        assert trc.translations['en']['send'] == 'Send'

    def test_tree_node_set_ancestor(self):
        ancestor = TreeNode()
        child = TreeNode('', is_root=False, is_ancestor=False, ancestor=ancestor)
        assert child.ancestor == ancestor

    def test_treenode_repr(self):
        capturedOutput = StringIO()
        sys.stdout = capturedOutput
        line = "h1: 'test'"
        node = TreeNode(line)
        print(node)
        sys.stdout = sys.__stdout__
        assert line in capturedOutput.getvalue()

    def test_add_children(self):
        root = Node('root', local_context={}, is_root=True, is_ancestor=True)
        root.document = self
        source = \
"""
div:
  p: 'test'
    p: 'test'
"""
        root.add_children([Egg(line) for line in source.splitlines() if line.strip()])

    def test_context_item_model_accessor(self):
        code = \
"""
+post:
  .title:
    &label: 'Title'
    &type: 'str'
#data: +post
"""
        zml.render(code)

    def test_field_meta_data(self):
        code = \
"""
+post:
  .title:
    &label: 'Title'
    &type: 'str'
p: +post.title&type
"""
        zml.render(code)

    def xtest_code_if_statement(self):
        code = \
"""
%import components
%inherit base

#user_active: True
#user_deleted: False

*content:
  %if user_active:
    p: 'User is active'
  %if user_deleted:
    p: 'User is deleted'
"""
        target = \
"""
<html>
  <head>
    <title>zml</title>
  </head>
  <body>
    <p>User is active
    </p>
  </body>
</html>
"""
        run_test(code, target)

    def test_load_document(self):
        os.chdir(os.path.dirname(__file__))
        filename = 'load.zml'
        result = zml.load(filename)
        data_dict = {'person': {'first_name': 'Melvin', 'last_name': 'Frohike', 'profession': 'Developer'}}
        assert result.local_context == data_dict

    def test_accessors(self):
        dirs = AppDirs("zml_client", "zml", version="1.0")
        data_directory = os.path.join(dirs.user_data_dir, 'data')
        try:
            os.makedirs(data_directory)
        except FileExistsError:
            pass
        zml.util.start_ipfs_daemon(data_directory)
        os.chdir(os.path.dirname(__file__))
        filename = 'data.zml'
        doc = zml.load(filename)
        value = doc.get('#person.first_name')
        value = doc.get('#person.more.deeper.structure')
        assert value == 'Richard'

    def test_model(self):
        os.chdir(os.path.dirname(__file__))
        filename = 'person.zml'
        doc = zml.load(filename)
        assert isinstance(doc.models['person'], Model)

    def test_something(self):
        os.chdir(os.path.dirname(__file__))
        filename = 'external_data.zml'
        doc = zml.load(filename)
        value = doc.get('#first_name')
        assert value == 'Richard'

    def test_rest(self):
        code = \
"""
%import components
%inherit base

@db: 'en.wikipedia.org/w/api.php?action=query&list=search&format=json&srsearch=rest'

#pages: @db

*content:
  %for p in #pages.query.search:
    p: '{#p.title}'
"""
        out = zml.render(code)
        # todo: more detailed test for rest
        assert 'Representational state transfer' in out

    def test_dispatch(self):
        code = \
"""
%import components

~main:
  view: '/view/{id}'
  edit: '/edit/{id}'

*view:
  h1: 'View'
  ul:
    li:
      base-linkto action='view' id=1: 'View'
    li:
      base-linkto action='edit' id=1: 'Edit'

*edit:
  h1: 'Edit'
  ul:
    li:
      base-linkto action='view' id=1: 'View'
    li:
      base-linkto action='edit' id=1: 'Edit'

html:
  head:
    title: 'zml'
  body:
    |main
"""
        target_view = \
"""
<html>
  <head>
    <title>zml</title>
  </head>
  <body>
      <h1>View</h1>
      <ul>
        <li>
          <a href="/view/1">View</a>
        </li>
        <li>
          <a href="/edit/1">Edit</a>
        </li>
      </ul>
  </body>
</html>
"""
        target_edit = \
"""
<html>
  <head>
    <title>zml</title>
  </head>
  <body>
      <h1>Edit</h1>
      <ul>
        <li>
          <a href="/view/1">View</a>
        </li>
        <li>
          <a href="/edit/1">Edit</a>
        </li>
      </ul>
  </body>
</html>
"""
        out = zml.render(code, path='/view/1')
        assert minimise(out) == minimise(target_view)
        out = zml.render(code, path='/edit/1')
        assert minimise(out) == minimise(target_edit)

    def test_import_external_component(self):
        code = \
"""
@base: 'QmbUgMuUAfmw8uaWuEhzGJCoGLy7ytN5Rb73XnVmmv5auh'

%import @base

~main:
  list: '/blog/posts'

base-linkto action='list': 'List'

"""
        target = \
"""
<a href="/blog/posts">List</a>\n
"""
        run_test(code, target)

    def test_render_external_component(self):
        response = zml.render_address('QmRP3FtcmmCvQ4sVcJLuLY9CipxmBCLpJp5B52GRi2Nhdd')
        print(response)

    def test_render_external_component_and_inherit(self):
        response = zml.render_address('QmQitcFTeRzDp6M7EPepzo6pbMnxuCEFaMZmUKw3GS4vfC', path='/blog/posts')
        print(response)

    def test_multiline(self):
        key = \
"""
----BEGIN PGP MESSAGE-----

jA0EBwMCKFOWDIApgLLx0o8BOb85gzkxIdVAE3tSIX9R/3yXthBUd5QPemx1Lfiz
pHpjmG/DOKJ1aN9ZwqzksAlgqLTf8UPRG9Ch/MPZoy9Q1R5KJv6QKlMPbn5XHqqo
NW5jSV5g2bX6pcl1FUqbCI9yfyDCw99Rxap01qWXxmlkD7uTp5tL2CFmg3SlDVKb
hAX8YpCjSYNDKlXL56O6rg==
=0C/y
-----END PGP MESSAGE-----
"""
        code = "#ciphertext: '{}'".format(key)
        document = zml.import_source(code)
        assert document.local_context['ciphertext'] == key



from zml.backend import IpfsBackend, RestBackend


class Resource:

    def __init__(self, address=None, lazy=True, code=None, mime_type=None):
        self.address = address
        self.code = code
        self.mime_type = mime_type
        self.is_media = False
        self.lazy = lazy
        self.loaded = False
        if not lazy:
            self.load()

    def load(self):
        if self.address[0] == 'Q':
            backend = IpfsBackend()
            self.type = 'ipfs'
            raw, mime_type = backend.load(self.address)
        elif self.address.startswith('http://'):
            backend = FileBackend()
            self.type = 'file'
            raw, mime_type = backend.load(self.address)
        else:
            backend = RestBackend()
            self.type = 'rest'
            raw, mime_type = backend.load(self.address)

        if mime_type in ['image/jpeg', 'image/png', 'image/gif']:
            self.code = raw
            self.is_media = True
        else:
            self.code = raw
            if not isinstance(raw, str) and not isinstance(raw, dict):
                self.code = self.code.decode('utf-8')
        self.mime_type = mime_type
        self.loaded = True

    def import_resource(self, document):
        if not self.loaded:
            self.load()
        if self.type in ['ipfs', 'file']:
            if self.mime_type in ['image/jpeg', 'image/png', 'image/gif']:
                raise Exception('Cannot import media files')
            else:
                return document.import_source(self.code)
        else:
            # update local context with json from rest document
            document.local_context.update(self.code)



from zml.node import Model


class Operator(object):
    pass


class Processor(Operator):

    def __init__(self, render=False, lazy=False):
        self.lazy = lazy

    def subtree_handler_default(self, node):
        for child in node.children:
            child.process_subtree(self)
        return ''

    def process(self, node):
        node.process(lazy=self.lazy)
        self.subtree_handler_default(node)

    def subtree_handler_for_loop(self, node):
        # prepare
        # iterator_descriptor = node.tokens['iterator']
        variable_descriptor = node.tokens['variable']
        if not self.lazy:
            iterator = node.eval_context_item(node.tokens)
            if isinstance(iterator, Model):
                iterator = iterator.values()
            for item in iterator:
                node.set_var(variable_descriptor, item)
                # execute
                self.subtree_handler_default(node)
#        else:
#            self.subtree_handler_default(node)

    def subtree_handler_if_statement(self, node):
        # prepare
        condition = node.eval_context_item(node.tokens)
        if not self.lazy:
            if condition:
                # execute
                self.subtree_handler_default(node)
        else:
            self.subtree_handler_default(node)


class Renderer(Operator):

    def __init__(self, render=False, lazy=False, base_render_level=0):
        self.base_render_level = base_render_level
        self.lazy = lazy

    def subtree_handler_default(self, node):
        out = ''
        for child in node.children:
            # prevent rendering of ComponentNode
            # ComponentNodes will be rendered explicitly
            if not child.is_component:
                out += child.render_subtree(self)
        return out

    def render(self, node):
        node.render()
        out = ''
        out += node.leading
        out += self.subtree_handler_default(node)
        out += node.trailing
        return out

    def subtree_handler_for_loop(self, node):
        # prepare
        out = ''
        if not self.lazy:
            variable_descriptor = node.tokens['variable']
            iterator = node.eval_context_item(node.tokens)
            if isinstance(iterator, Model):
                iterator = iterator.fields.values()
            for item in iterator:
                node.set_var(variable_descriptor, item)
                # execute
                out += self.subtree_handler_default(node)
        else:
            out += self.subtree_handler_default(node)
        return out

    def subtree_handler_if_statement(self, node):
        # prepare
        out = ''
        if not self.lazy:
            condition = node.eval_context_item(node.tokens)
            if condition:
                # execute
                out += self.subtree_handler_default(node)
        else:
            out += self.subtree_handler_default(node)
        return out



import os
import time
from pprint import pprint
from zml.exceptions import FileNotLoadedException, DocumentNotDefinedException
import subprocess
import shlex
import socket


def load_file(filename):
    try:
        with open(filename, 'r', encoding='utf-8') as f:
            content = f.read()
            return content
    except Exception:
        raise FileNotLoadedException


def find_file_in_dirs(filename, directories):
    for d in directories:
        abs_path = os.path.join(d, filename)
        if os.path.exists(abs_path):
            return abs_path
    raise DocumentNotDefinedException


def minimise(code):
    # return code.replace('\n', '')
    return code.strip()


def deb(obj):
    pprint(obj.asDict())


def set_base_path(base_path):
    os.chdir(base_path)


def start_ipfs_daemon(user_data_dir=None):
    if not user_data_dir:
        raise Exception('No user data dir defined.')
    daemon_is_running = False
    while True:
        # rc, std_out, std_err = run_command('ipfs daemon --enable-namesys-pubsub --enable-pubsub-experiment')
        rc, std_out, std_err = run_command('ipfs daemon')
        token_out = 'Daemon is ready'
        token_err = 'ipfs daemon is running'
        token_not_inited = 'no IPFS repo found'
        if std_out and any(token in token_out for token in std_out.decode('utf-8')):
            daemon_is_running = True
        if std_err and any(token in token_err for token in std_err.decode()):
            daemon_is_running = True
        if std_err and any(token in token_not_inited for token in std_err.decode()):
            os.chdir(user_data_dir)
            rc, std_out, std_err = run_command('ipfs init')
        if daemon_is_running:
            # check if daemon is listening on port 5001
            sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
            result = sock.connect_ex(('127.0.0.1', 5001))
            if result == 0:
                break
        time.sleep(1)


def run_command(command):
    process = subprocess.Popen(shlex.split(command), stdout=subprocess.PIPE, stderr=subprocess.PIPE)
    try:
        std_out, std_err = process.communicate(timeout=1)
    except subprocess.TimeoutExpired:
        std_out, std_err = None, None
    return process.returncode, std_out, std_err



import gnupg
try:
    import oqs
except:
    pass
from Crypto.Cipher import AES
from Crypto import Random
from pkcs7 import PKCS7Encoder
import zml


def generate_keypair():

    with oqs.KeyEncapsulation(kemalg) as client:
        with oqs.KeyEncapsulation(kemalg) as server:
            print("\nStarting key encapsulation")
            pprint(client.details)

            # client generates its keypair
            public_key = client.generate_keypair()
            # optionally, the secret key can be obtained by calling export_secret_key()
            # and the client can later be reinstantiated with the key pair:
            # secret_key = client.export_secret_key()
            # store key pair, wait... (session resumption):
            # client = oqs.OQS_KEM(kemalg, secret_key)

            # the server encapsulates its secret using the client's public key
            ciphertext, shared_secret_server = server.encap_secret(public_key)

            # the client decapsulates the the server's ciphertext to obtain the shared secret
        # shared_secret_client = client.decap_secret(ciphertext)


def generate_secret_key_aes():
    rndfile = Random.new()
    key = rndfile.read(32)
    return key


def store_secret_key():
    pass


def load_secret_key(keys_directory, key_file):
    gpg = gnupg.GPG(gnupghome=keys_directory)
    document = zml.import_file(key_file)
    encrypted_key = document.local_context['ciphertext']
    decryption = gpg.decrypt(encrypted_key)
    if decryption.status == 'decryption ok':
        decrypted_key = decryption.data
    else:
        raise Exception('Decryption failure')
    return decrypted_key

# gpg = gnupg.GPG(gnupghome=keys_directory, verbose=True)
# gpg.encoding = 'utf-8'
# input_data = gpg.gen_key_input(
#     name_email='testgpguser@mydomain.com',
#     passphrase=pass_phrase)


def create_secret_key_file(keys_directory, key_file):
    key = generate_secret_key_aes()
    gpg = gnupg.GPG(gnupghome=keys_directory)
    encrypted_key = gpg.encrypt(key, None, symmetric=True)
    code = "&cipher: 'AES-256'\n#ciphertext: '\n{}'".format(encrypted_key)
    with open(key_file, 'w', encoding='utf-8') as f:
        f.write(code)


def encrypt(code, key):
    cipher = encrypt_message(code, key)
    return cipher


def decrypt(cipher, key):
    iv = 'skeurTFGHU384756'
    obj2 = AES.new(key, AES.MODE_CBC, iv)
    decoded_message = obj2.decrypt(cipher).decode('utf-8')
    return decoded_message


def encrypt_message(message, key):
    print(key)
    # key = self.generate_secret_key_aes()
    iv = 'skeurTFGHU384756'
    obj = AES.new(key, AES.MODE_CBC, iv)
    encoder = PKCS7Encoder()
    pad_message = encoder.encode(message)
    cipher = obj.encrypt(pad_message)
    return cipher



from zml.exceptions import VariableNotDefinedException
from zml.semantic import *


class RenderingContext(object):

    def get_var(self, variable_descriptor):
        if variable_descriptor in self.local_context:
            return self.local_context[variable_descriptor]
        if variable_descriptor in self.global_context:
            return self.global_context[variable_descriptor]
        raise VariableNotDefinedException('The variable {} is not defined'.format(variable_descriptor))

    def set_var(self, variable_descriptor, value):
        self.local_context[variable_descriptor] = value




import math
from pyparsing import *
import operator


exprStack = []

variables = {}


def pushFirst(strg, loc, toks):
    exprStack.append(toks[0])


def pushUMinus(strg, loc, toks):
    for t in toks:
        if t == '-':
            exprStack.append('unary -')
            # ~ exprStack.append( '-1' )
            # ~ exprStack.append( '*' )
        else:
            break


bnf = None


def BNF():
    """
    expop   :: '^'
    multop  :: '*' | '/'
    addop   :: '+' | '-'
    integer :: ['+' | '-'] '0'..'9'+
    atom    :: PI | E | real | fn '(' expr ')' | '(' expr ')'
    factor  :: atom [ expop factor ]*
    term    :: factor [ multop factor ]*
    expr    :: term [ addop term ]*
    """
    global bnf
    if not bnf:
        # point = Literal(".")
        e = CaselessLiteral("E")
        # ~ fnumber = Combine( Word( "+-"+nums, nums ) +
        # ~ Optional( point + Optional( Word( nums ) ) ) +
        # ~ Optional( e + Word( "+-"+nums, nums ) ) )
        fnumber = Regex(r"[+-]?\d+(:?\.\d*)?(:?[eE][+-]?\d+)?")
        ident = Word(alphas, alphas + nums + "_$")

        plus = Literal("+")
        minus = Literal("-")
        mult = Literal("*")
        div = Literal("/")
        lpar = Literal("(").suppress()
        rpar = Literal(")").suppress()
        addop = plus | minus
        multop = mult | div
        expop = Literal("^")
        pi = CaselessLiteral("PI")

        expr = Forward()
        atom = ((0, None) * minus + (pi | e | fnumber | ident + lpar + expr + rpar | ident).setParseAction(pushFirst) | Group(lpar + expr + rpar)).setParseAction(pushUMinus)

        # by defining exponentiation as "atom [ ^ factor ]..." instead of "atom [ ^ atom ]...", we get right-to-left
        # exponents, instead of left-to-right
        # that is, 2^3^2 = 2^(3^2), not (2^3)^2.
        factor = Forward()
        factor << atom + ZeroOrMore((expop + factor).setParseAction(pushFirst))

        term = factor + ZeroOrMore((multop + factor).setParseAction(pushFirst))
        expr << term + ZeroOrMore((addop + term).setParseAction(pushFirst))
        bnf = expr
    return bnf


# map operator symbols to corresponding arithmetic operations
epsilon = 1e-12
opn = {"+": operator.add,
       "-": operator.sub,
       "*": operator.mul,
       "/": operator.truediv,
       "^": operator.pow}
fn = {"sin": math.sin,
      "cos": math.cos,
      "tan": math.tan,
      "abs": abs,
      "trunc": lambda a: int(a),
      "round": round,
      "sgn": lambda a: abs(a) > epsilon and cmp(a, 0) or 0}


def evaluateStack(s):
    op = s.pop()
    if op == 'unary -':
        return -evaluateStack(s)
    if op in "+-*/^":
        op2 = evaluateStack(s)
        op1 = evaluateStack(s)
        return opn[op](op1, op2)
    elif op == "PI":
        return math.pi  # 3.1415926535
    elif op == "E":
        return math.e  # 2.718281828
    elif op in fn:
        return fn[op](evaluateStack(s))
    elif op[0].isalpha():
        if op in variables:
            return variables[op]
        raise Exception("invalid identifier '%s'" % op)
    elif '.' in op:
        return float(op)
    else:
        return int(op)


def named(marker):
    def parse_action_impl(s, l, t):
        return (marker, t)
    return parse_action_impl


void_elements = [
    '', 'area', 'base', 'br', 'col', 'command',
    'embed', 'hr', 'img', 'input', 'keygen',
    'link', 'meta', 'param', 'source', 'track', 'wbr'
]

inline_elements = [
    'h1', 'a', 'b', 'strong', 'i', 'em', 'title', 'label', 'button'
]


# general node semantics
ARITH_EXPRESSION = BNF()
COLON = Literal(':')

ALL = Word(printables + ' ', excludeChars=":")('all')

UNQUOTED_WORDS = Word(printables + ' ', excludeChars="'{}<>")

DESCRIPTOR = Word(alphanums + '_')
NAMESPACE_DESCRIPTOR = DESCRIPTOR('namespace') + '-' + DESCRIPTOR('namespaced_descriptor')
GLYPH = Word('*$&~!+%#@|', max=1)('glyph')

DESCRIPTOR_LINE = Optional(GLYPH)('glyph') + DESCRIPTOR('descriptor')
NAMESPACE_DESCRIPTOR_LINE = Optional(GLYPH)('glyph') + NAMESPACE_DESCRIPTOR('namespace_descriptor')
EMPTY = Empty()

LIST_ITEM_LINE = Literal('-')('list_item_glyph')
LINE = Optional(LIST_ITEM_LINE | NAMESPACE_DESCRIPTOR_LINE | DESCRIPTOR_LINE | EMPTY('descriptor')) + Optional(COLON)

NUMBER = Word(nums)
QUOTED_STRING = QuotedString(quoteChar="'")
TRANSLATION = Suppress('!') + DESCRIPTOR('language_code')
PROPERTY_ACCESSOR = Suppress('.') + Word(alphanums + '_')
META_ACCESSOR = Suppress('&') + Word(alphanums + '_')
TRANSLATION_ACCESSOR = Suppress('!') + DESCRIPTOR('translation_descriptor') + ZeroOrMore(PROPERTY_ACCESSOR.setParseAction(named('property')))('properties')
MODEL = Literal('+')('glyph') + DESCRIPTOR('model_descriptor')
FIELD = Optional(Literal('.')('glyph')) + DESCRIPTOR('field_descriptor')
COMPONENT_DESCRIPTOR = Suppress('*') + DESCRIPTOR('component_descriptor')
CONTEXT_ACCESSOR = Literal('_context')('context_accessor')
CONTEXT_ITEM = Optional('.') + DESCRIPTOR('context_item')

UID = Suppress('#') + Word(alphanums, excludeChars='# : .')
CLS = Suppress('.') + Word(alphanums, excludeChars='. : #')
CLASSES = ZeroOrMore(CLS)('classes')
ID_CLASSES = Optional(UID)('uid') + CLASSES
ATTRIBUTE_KEY = Word(alphanums + '_' + '-', excludeChars=": = '")
# PARAM = (CONTEXT_ITEM('context_item') | CONTEXT('context')) + Optional(',')
# PARAM_LIST = ZeroOrMore(PARAM)
# FRAMEWORK_COMPONENT = Suppress('*core-') + DESCRIPTOR('framework_component') + PARAM_LIST('param_list')
VALUE_ACCESSOR = Literal('_value')('value_accessor')
LIST = Literal('[]')('list')
DICT = Literal('{}')('dict')
BOOLEAN = (Literal('True') or Literal('False'))('boolean')
LITERAL = (quotedString.addParseAction(removeQuotes) ^ pyparsing_common.number)('literal')
EXPRESSION = Forward()

code = \
"""
#ciphertext: '
----BEGIN PGP MESSAGE-----

jA0EBwMCKFOWDIApgLLx0o8BOb85gzkxIdVAE3tSIX9R/3yXthBUd5QPemx1Lfiz
pHpjmG/DOKJ1aN9ZwqzksAlgqLTf8UPRG9Ch/MPZoy9Q1R5KJv6QKlMPbn5XHqqo
NW5jSV5g2bX6pcl1FUqbCI9yfyDCw99Rxap01qWXxmlkD7uTp5tL2CFmg3SlDVKb
hAX8YpCjSYNDKlXL56O6rg==
=0C/y
-----END PGP MESSAGE-----
'
"""


# ParserElement.setDefaultWhitespaceChars(" \t")
INLINE_CONTENT = Forward()

DATA_NODE_ACCESSOR = Literal('#')('glyph') + DESCRIPTOR('data_node_descriptor')
SLOT_NODE = Literal('|')('glyph') + DESCRIPTOR('descriptor')
SLOT_NODE_INLINE = Literal('|')('glyph') + DESCRIPTOR('descriptor')
RESOURCE_ACCESSOR = Suppress('@') + DESCRIPTOR('resource')
COMBINED_ACCESSOR = Optional(RESOURCE_ACCESSOR) + (DATA_NODE_ACCESSOR | MODEL | CONTEXT_ITEM) + ZeroOrMore(PROPERTY_ACCESSOR.setParseAction(named('property')) | META_ACCESSOR.setParseAction(named('meta_data')))('properties')  # noqa
CID = Combine(Literal('Q') + DESCRIPTOR)
SEGMENT = (DATA_NODE_ACCESSOR |
           MODEL |
           META_ACCESSOR |
           CONTEXT_ITEM |
           OneOrMore(PROPERTY_ACCESSOR.setParseAction(named('property')))('properties'))
FULL_PATH = Optional(CID('resource')) + OneOrMore(SEGMENT.setParseAction(named('segment')))('segments')  # noqa
MOUSTACHE_EXPRESSION = CONTEXT_ACCESSOR | COMBINED_ACCESSOR | CONTEXT_ITEM | COMPONENT_DESCRIPTOR
MOUSTACHE_ATTRIBUTE_VALUE = NUMBER('number') | QUOTED_STRING('quoted_string') | MOUSTACHE_EXPRESSION
MOUSTACHE_ATTRIBUTE = ATTRIBUTE_KEY('moustache_attribute_key') + Optional(Literal('=') + MOUSTACHE_ATTRIBUTE_VALUE('moustache_attribute_value'))
MOUSTACHE_ATTRIBUTES = ZeroOrMore(Group(MOUSTACHE_ATTRIBUTE('moustache_attribute')))
FRAMEWORK_COMPONENT = Suppress('*core-') + DESCRIPTOR('framework_component') + MOUSTACHE_ATTRIBUTES('moustache_attributes')
MOUSTACHE = Suppress('{')
MOUSTACHE += EXPRESSION
MOUSTACHE += Suppress('}')
ATTRIBUTE_VALUE = Optional(Literal("'") | Literal('"')) + MOUSTACHE('moustache') + Optional(Literal("'") | Literal('"')) | NUMBER('number') | QUOTED_STRING('quoted_string') | EXPRESSION
ATTRIBUTE = ATTRIBUTE_KEY('attribute_key') + Optional(Literal('=') + ATTRIBUTE_VALUE('attribute_value'))
ATTRIBUTES = ZeroOrMore(Group(ATTRIBUTE('attribute')))
EXPRESSION << (
    BOOLEAN ^ LITERAL ^ LIST ^ DICT ^ VALUE_ACCESSOR ^ FRAMEWORK_COMPONENT
    ^ RESOURCE_ACCESSOR
    ^ COMBINED_ACCESSOR('context_item_with_property') ^ CONTEXT_ITEM
    ^ MODEL ^ COMPONENT_DESCRIPTOR ^ ARITH_EXPRESSION('arith_expression')
    ^ TRANSLATION_ACCESSOR('translation_accessor'))

INLINE_SEMANTICS_CONTENT = ZeroOrMore(UNQUOTED_WORDS.leaveWhitespace()('inline_semantic_content_words') | TRANSLATION_ACCESSOR('translation_accessor') | MOUSTACHE('moustache'))
INLINE_SEMANTICS_ELEMENT = DESCRIPTOR('descriptor') + Optional(' ') + ID_CLASSES('id_classes') + Optional(' ')
INLINE_SEMANTICS_ELEMENT += ATTRIBUTES('attributes') + Optional(' ')
INLINE_SEMANTICS_ELEMENT += Optional(COLON('colon') + Suppress(Optional(' ')) +
                                     INLINE_SEMANTICS_CONTENT('inline_semantics_content'))
INLINE_SEMANTICS = Suppress('<') + INLINE_SEMANTICS_ELEMENT + Optional(' ') + Suppress('>') + Optional(' ')('trailing_space')

INLINE_CONTENT << \
    (Suppress("'") + ZeroOrMore(LineEnd()).setParseAction(named('inline_content_newlines')) + ZeroOrMore((Literal(' ')('space')
    ^ INLINE_SEMANTICS.leaveWhitespace().setParseAction(named('inline_semantics'))
    ^ UNQUOTED_WORDS.setParseAction(named('inline_content_words'))
    ^ MOUSTACHE.setParseAction(named('moustache'))) + ZeroOrMore(LineEnd()).setParseAction(named('inline_content_newlines'))) + Suppress("'")
    ) | SLOT_NODE_INLINE('slot_node')

ELEMENT = Optional(DESCRIPTOR('descriptor')) + ID_CLASSES('id_classes')
ELEMENT += ATTRIBUTES('attributes') + Optional(COLON)('colon') + Optional(' ')
ELEMENT += Optional(INLINE_CONTENT.leaveWhitespace()('inline_content') ^ EXPRESSION('expression'))

# instruction semantics
FILENAME = DESCRIPTOR('filename')
URL = Word(printables, excludeChars='* "  : # { }' + "'")

RESOURCE = Suppress('@') + DESCRIPTOR('resource') + COLON + "'" + (URL | DESCRIPTOR)('source') + "'"

INHERIT = Suppress('%') + Literal('inherit')('instruction')
INHERIT += FILENAME('documentfile') | RESOURCE_ACCESSOR

IMPORT = Suppress('%') + Literal('import')('instruction')
IMPORT += FILENAME('documentfile') | RESOURCE_ACCESSOR('resource_accessor')

INCLUDE = Suppress('%') + Literal('include')('instruction')
INCLUDE += FILENAME('documentfile')

NAMESPACE = Suppress('%') + Literal('namespace')('instruction')
NAMESPACE += DESCRIPTOR('namespace') + '=' + URL

INSTRUCTION = INHERIT | IMPORT | INCLUDE | NAMESPACE


# route semantics
url_chars = alphanums + '-_.~%+'
PATH_SEGMENT = Word(url_chars)
PATH_VARIABLE = Suppress('{') + DESCRIPTOR('variable_descriptor') + Suppress('}')
PATH_ITEM = Suppress(Optional('/'))('leading_slash') + (PATH_SEGMENT.setParseAction(named('path_segment'))
                         | PATH_VARIABLE.setParseAction(named('path_variable')))
PATH_ITEMS = ZeroOrMore(PATH_ITEM('path_item'))
ROUTE_PATH = PATH_ITEMS('path_items')
ROUTE_PATH += Optional('/')('trailing_slash')
ROUTE = DESCRIPTOR('descriptor') + COLON + Optional(' ') + Suppress("'") + ROUTE_PATH('route_path') + Suppress("'")


# code semantics
SOMETHING = Suppress('%') + Word(printables + ' ')('some_thing')
ITERATOR = COMBINED_ACCESSOR | CONTEXT_ITEM
IF_STATEMENT = Suppress('%') + Suppress('if') + EXPRESSION('expression') + Suppress(':')
FOR_LOOP = Suppress('%') + Suppress('for') + DESCRIPTOR('variable') + Suppress('in')
FOR_LOOP += ITERATOR('iterator') + Suppress(':')
CODE = INHERIT | IMPORT | INCLUDE | NAMESPACE | IF_STATEMENT('if_statement') | FOR_LOOP('for_loop') | SOMETHING('some_thing')

COMPONENT_CALL_BODY = DESCRIPTOR('namespace') - Suppress('-') - DESCRIPTOR('component_descriptor')
COMPONENT_CALL = COMPONENT_CALL_BODY + ATTRIBUTES('attributes')
COMPONENT_CALL += Optional(COLON)('colon') + Optional(' ') + Optional(INLINE_CONTENT.leaveWhitespace()('inline_content') ^ EXPRESSION('expression'))
EGG = Optional(ALL('key')) + Optional(COLON)('colon') + Optional(ALL('value'))

# list items
LIST_ITEM = Literal('-')('list_item_glyph') + Optional(" ") + Optional(INLINE_CONTENT.leaveWhitespace())('inline_content')

ASSIGNMENT_VALUE = Optional(EXPRESSION('expression') ^ INLINE_CONTENT.leaveWhitespace()('inline_content'))
ASSIGNMENT = Optional(GLYPH) + DESCRIPTOR('descriptor') + COLON('colon') + Optional(' ') + ASSIGNMENT_VALUE('assignment_value')
# we have to definde DATA_NODE_VALUE containig INLINE_CONTENT here (not using Forward definition for INLINE_CONTENT,
# because using Forward definition seem to cause missing items in named label "inline_content" in parseResults
DATA_NODE_VALUE = Optional(EXPRESSION('expression') ^ INLINE_CONTENT.leaveWhitespace()('inline_content'))
DATA_NODE = Literal('#')('glyph') + DESCRIPTOR('data_descriptor') + COLON('colon') + Optional(' ') + DATA_NODE_VALUE('data_node_value')

# reintroduce newline to whitespace after last usage of INLINE_CONTENT, as INLINE_CONTENT is the only place where we use newlines in syntax
# ParserElement.setDefaultWhitespaceChars(" \n\t")

URL_PATH_VARIABLE = Suppress('{') + DESCRIPTOR('descriptor') + Suppress('}')
# URL_PATH = Suppress(Optional('/')) + ZeroOrMore((DESCRIPTOR.setParseAction(named('segment')) |
#                       URL_PATH_VARIABLE.setParseAction(named('variable'))) + Suppress(Optional('/')))('segments')
URL_PATH = Suppress(Optional('/')) + ZeroOrMore(DESCRIPTOR('segment') + Suppress(Optional('/')))('segments')
url_chars = alphanums + '-_.~%+'
fragment = Combine((Suppress('#') + Word(url_chars)))('fragment')
scheme = oneOf('http https ftp file')('scheme')
host = Combine(delimitedList(Word(url_chars), '.'))('host')
port = Suppress(':') + Word(nums)('port')
user_info = (Word(url_chars)('username') + Suppress(':') + Word(url_chars)('password') + Suppress('@'))

query_pair = Group(Word(url_chars) + Suppress('=') + Word(url_chars))
query = Group(Suppress('?') + delimitedList(query_pair, '&'))('query')

path = Combine(Suppress('/') + OneOrMore(~query + Word(url_chars + '/')))('path') + Optional(query) + Optional(fragment)

# url_parser = (scheme + Suppress('://') + Optional(user_info) + host + Optional(port) + Optional(path) + Optional(query) + Optional(fragment))
url_parser = (Optional(path) + Optional(query) + Optional(fragment))

BEGIN_MULTILINE = Combine(LINE + ' ' + Literal("'") + LineEnd())
END_MULTILINE = Literal("'") + LineEnd()



from zml.exceptions import (DocumentNotDefinedException, TranslationNotDefinedException)
from zml.context import RenderingContext
from zml.node import (Egg, EmptyNode, AssignmentNode, CodeNode, ComponentNode, ComponentCallNode,
                      DataNode, ModelNode, Node, MetaNode, ListItemNode, SlotNode, ResourceNode,
                      RouterNode, TranslationNode)
from zml.node import eval_context_item
from zml.backend import FileBackend, IpfsBackend
from zml.operator import Processor, Renderer
from zml.util import find_file_in_dirs, load_file
from zml.request import Request
from zml.resource import Resource
from zml.semantic import *
from zml.exceptions import NotFoundException


type_mapping = {
    '*': ComponentNode,
    '@': ResourceNode,
    '+': ModelNode,
    '~': RouterNode,
    '!': TranslationNode,
    '#': DataNode,
    '&': MetaNode,
    '%': CodeNode,
    '|': SlotNode,
    '-': ListItemNode
}


def render_address(address, path=None, get_params=None, post_params=None):
    backend = IpfsBackend()
    if address[0] == 'Q':
        code, mime_type = backend.load(address)
    out = render(code, path=path, get_params=get_params, post_params=post_params, cid=address)
    # out = Document(source=code, request=request).render()
    return out


def get_address(address):
    if address[0] == 'Q':
        backend = IpfsBackend()
        raw, mime_type = backend.load(address)
        if mime_type in ['image/jpeg', 'image/png', 'image/gif']:
            res = raw
        else:
            res = import_source(raw, cid=address)
    elif address[0] == 'http://':
        backend = RestBackend()
        res, mime_type = backend.load(address)
    else:
        backend = FileBackend()
        raw, mime_type = backend.load(address)
        if mime_type in ['image/jpeg', 'image/png', 'image/gif']:
            res = raw
        else:
            res = import_source(raw, cid=address)
    return res, mime_type


def get_path(path):
    res = FULL_PATH.parseString(path)
    if 'resource' in res:
        address = res['resource']
        backend = IpfsBackend()
        if address[0] == 'Q':
            source, mime_type = backend.load(address)
            document = import_source(source, cid=address)
            node = document.root
            if 'segments' in res:
                for segment in res['segments']:
                    node = node.eval_segment(segment)
                return node
    else:
        raise Exception("Content ID missing in path.")


def get_data_node(line, document, base_indent=0):
    res = LINE.parseString(line)
    if 'list_item_glyph' in res:
        return ListItemNode(line, document=document, global_context=document.global_context, base_indent=base_indent, renderer=Renderer())
    else:
        res = ASSIGNMENT.parseString(line)
        if 'descriptor' in res:
            return AssignmentNode(line, document=document, global_context=document.global_context,
                                  base_indent=base_indent, renderer=Renderer())


def get_document_node(line, document, base_indent=0):
    res = LINE.parseString(line)
    if 'descriptor' in res:
        if 'glyph' in res:
            glyph = res['glyph']
            if glyph in type_mapping:
                descriptor = res['descriptor']
                doc = type_mapping[glyph](line, document=document, is_ancestor=True, descriptor=descriptor,
                                          global_context=document.global_context, base_indent=base_indent,
                                          renderer=Renderer())
                return doc
        else:
            return Node(line, document=document, global_context=document.global_context, base_indent=base_indent,
                        renderer=Renderer())
    elif 'namespace_descriptor' in res:
        return ComponentCallNode(line, document=document, global_context=document.global_context,
                                 base_indent=base_indent, renderer=Renderer())
    else:
        return EmptyNode(line, document=document, global_context=document.global_context, base_indent=base_indent,
                         renderer=Renderer())


class DocumentRenderingContext(RenderingContext):

    def __init__(self, global_context={}, local_context={}, request=Request()):
        self.local_context = local_context
        self.router = dict()
        self.default_router = None
        self.request = request
        self.global_context = dict()
        self.global_context['_resources'] = dict()
        self.global_context['_translations'] = dict()
        self.global_context['_models'] = dict()
        if global_context is not None:
            self.global_context.update(global_context)
        self.namespaces = dict()
        self.namespaces['_default'] = dict()
        self.translations = dict()
        self.dispatched_views = dict()

    def get_translation(self, variable_descriptor, language=None):
        if language is None:
            try:
                language = next(iter(self.translations))
            except Exception:
                raise TranslationNotDefinedException
        if variable_descriptor in self.translations[language]:
            return self.translations[language][variable_descriptor]
        raise TranslationNotDefinedException

    def set_translation(self, language, variable_descriptor, value):
        if language not in self.translations:
            self.translations[language] = dict()
        self.translations[language][variable_descriptor] = value

    def dispatch_routes(self, path):
        res = URL_PATH.parseString(path)
        path_variables = dict()
        matched_view = None
        if not self.router:
            # no routers defined, we cannot dispatch and return without exception
            return
        for router_name in self.router:
            for view_name in self.router[router_name]:
                stop_view_check = False
                url_path_items = res['segments']
                router_path_items = self.router[router_name][view_name]
                if url_path_items is None:
                    url_path_items = []
                if router_path_items is None:
                    router_path_items = []
                if len(url_path_items) != len(router_path_items):
                    # path is longer than route definition, match not possible
                    stop_view_check = True
                    continue
                for i, url_path_item in enumerate(url_path_items):
                    router_path_item = router_path_items[i]
                    if router_path_item[0] == 'path_segment':
                        # segment
                        route_segment = router_path_item[1][0]
                        if route_segment != url_path_item:
                            stop_view_check = True
                            continue
                    elif router_path_item[0] == 'path_variable':
                        # variable
                        path_variables[router_path_item[1][0]] = url_path_item
                if stop_view_check:
                    continue
                # all segments matched
                matched_view = view_name
                self.dispatched_views[router_name] = matched_view
                return matched_view
        raise NotFoundException('Not found')


class Document(DocumentRenderingContext):

    def __init__(self, filename=None, source=None, lookup=None,
                 viewhelperdir=None, base_indent=0, namespaces={},
                 inherited_document=None, request=Request(), cid=None,
                 *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.filename = filename
        self.source = source
        self.cid = cid
        self.request = request
        self.lookup = lookup
        self.base_indent = base_indent
        self.namespace = '_default'
        self.namespaces = namespaces
        self.models = dict()
        self.resources = dict()
        self.inheriting_document = None
        self.inherited_document = inherited_document
        self.viewhelperdir = viewhelperdir
        self.localpage = dict()
        self.namespacemode = False
        self.language = None
        self.raw_mode = False

    def render(self,
               filename=None, source=None,
               local_context=None, global_context=None,
               base_indent=None):
        if local_context is None:
            local_context = {}
        if base_indent is None:
            base_indent = self.base_indent
        zmlsource = None
        if filename:
            document_path = self.get_absolute_path(filename)
            zmlsource = load_file(document_path)
        elif source:
            zmlsource = source
        elif self.filename:
            document_path = self.get_absolute_path(self.filename)
            zmlsource = load_file(document_path)
        elif self.source:
            zmlsource = self.source
        if zmlsource:
            result = self.render_source(source=zmlsource, local_context=self.local_context,
                                        request=self.request, base_indent=base_indent)
            return result
        else:
            raise DocumentNotDefinedException

    def get_source_from_file(self, filename):
        zmlsource = None
        if filename:
            document_path = self.get_absolute_path(filename)
            zmlsource = load_file(document_path)
        elif self.filename:
            document_path = self.get_absolute_path(self.filename)
            zmlsource = load_file(document_path)
        return zmlsource

    # todo: eventually rename to load, design loading concept for zml
    def import_document(self,
                        filename=None, source=None,
                        request=None, base_indent=None):
        """
        Imports a document in this document.
        """
        if base_indent is None:
            base_indent = self.base_indent
        zmlsource = self.get_source_from_file(filename)
        imported_document = Document(source=zmlsource, local_context=self.local_context,
                                     global_context=self.global_context)
        # todo: decide, if we have to decouple imported document from router setup and contexts of importing document
        # imported_document.router = self.router
        # imported_document.default_router = self.default_router
        if zmlsource:
            imported_document = self
            # todo: refactor, so no namespace switch is necessary
            # save namespace of importing document
            namespace_of_importing_document = self.namespace
            imported_document.translations = self.translations
            imported_document.import_source(zmlsource, request=request, base_indent=base_indent)
            self.namespace = namespace_of_importing_document
            self.namespaces.update(imported_document.namespaces)
            self.resources.update(imported_document.resources)
            self.models.update(imported_document.models)
            if not self.source:
                self.source = zmlsource
        else:
            raise DocumentNotDefinedException

    def get_absolute_path(self, filename):
        if filename is None:
            raise DocumentNotDefinedException
        if self.lookup is None:
            self.lookup = DocumentLookup(['.'])
        absolute_path = self.lookup.get_absolute_path(filename)
        return absolute_path

    def create_node_from_egg(self, egg, parent):
        # set parent, so is_data can check data context
        egg.parent = parent
        # check if parent forces children type, otherwise set type by parsing line
        if egg.line == '':
            # newline always creates empty node
            node = EmptyNode(egg.line, document=self)
        elif parent.get_children_type():
            node = parent.get_children_type()(egg.line, document=self)
        else:
            if parent.is_data() or isinstance(parent, AssignmentNode):
                node = get_data_node(egg.line, document=self)
            else:
                node = get_document_node(egg.line, document=self)
        node.children = egg.children
        node.body = egg.body
        node.link = egg.link
        return node

    def set_relations(self, node, ancestor, render_level=-1):
        node.render_level = render_level
        if not node.line.startswith('%') and not node.line.startswith('*'):
            render_level += 1
        if node.line.startswith('*'):
            node.is_ancestor = True
            node.local_context = node.local_context = node.document.local_context
            ancestor = node
        elif node.line.startswith('#'):
            node.local_context = node.ancestor.local_context
            node.value = None
            node.document.local_context[node.descriptor] = node.value
            ancestor = node.ancestor
        else:
            ancestor = node.ancestor
            node.local_context = node.ancestor.local_context
        for i, child in enumerate(node.children):
            child_node = self.create_node_from_egg(child, node)
            node.children[i] = child_node
            if isinstance(child_node, DataNode):
                node.data[child_node.descriptor] = node.children[i]
            node.children[i].parent = node
            node.children[i].ancestor = ancestor
            node.children[i].minimise = node.minimise
            self.set_relations(node.children[i], ancestor=node, render_level=render_level)

    def source_to_tree(self, source, base_indent=0):
        root = Node('root', local_context=self.local_context, is_root=True, is_ancestor=True, base_indent=base_indent, renderer=Renderer())
        root.document = self
        children = list()
        multiline = False
        for line_number, line in enumerate(source.splitlines()):
            try:
                BEGIN_MULTILINE.parseString(line)
                multiline = True
                multiline_code = line
            except Exception:
                if multiline:
                    multiline_code += "\n" + line
                else:
                    children.append(Egg(line, line_number=line_number+1, base_indent=base_indent))
            try:
                END_MULTILINE.parseString(line)
                multiline = False
                children.append(Egg(multiline_code, line_number=line_number+1, base_indent=base_indent))
            except Exception:
                pass
        root.add_children(children)
        self.set_relations(root, ancestor=root)
        return root

    def render_source(self, source=None,
                    local_context={},
                    indent_global='',
                    source_indent_level=0,
                    local_context_item='_root',
                    global_context=None,
                    request=Request(),
                    base_indent=None):
        if base_indent is None:
            base_indent = self.base_indent
        root = self.source_to_tree(source, base_indent=base_indent)
        if self.request.path:
            self.dispatch_routes(self.request.path)
        # we have to process the subtree to retrieve the routes
        # which are required by the components using the dispatcher
        root.process_subtree(Processor(lazy=True))
        out = root.render_subtree(Renderer())
        if self.inheriting_document:
            # here we set local_context of inheriting document to local_context of inherited document
            # todo: decide how we design the scopes of contexts
            if isinstance(self.inheriting_document, Resource):
                self.inheriting_document.load()
                inheriting_document = Document(source=self.inheriting_document.code, inherited_document=self,
                                               request=request)
                inheriting_document.dispatched_views = self.dispatched_views
                out = inheriting_document.render(local_context=self.local_context,
                                                 base_indent=base_indent)
            else:
                inheriting_document = Document(self.inheriting_document, inherited_document=self, request=request)
                inheriting_document.dispatched_views = self.dispatched_views
                out = inheriting_document.render(local_context=self.local_context,
                                                 base_indent=base_indent)
        return out

    def import_source(self, source=None,
                    indent_global='',
                    source_indent_level=0,
                    local_context_item='_root',
                    request=Request(),
                    base_indent=None):
        """
        Imports source into document.
        """
        # if no source is given the templates source is used as fallback
        if source is None:
            source = self.source

        namespace_of_importing_document = self.namespace
        if base_indent is None:
            base_indent = self.base_indent
        root = self.source_to_tree(source, base_indent=base_indent)
        if self.request and self.request.path:
            self.dispatch_routes(self.request.path)
        root.process_subtree(Processor(lazy=True))
        self.namespace = namespace_of_importing_document
        return root

    def get_language(self, language):
        return self.language

    def set_language(self, language):
        self.language = language

    def get(self, path):
        res = COMBINED_ACCESSOR.parseString(path)
        value = eval_context_item(res, self)
        return value


class DocumentLookup(object):

    def __init__(self, directories=None,
                 module_directory=None, input_encoding=None):
        self.directories = directories

    def get_document(self, filename, local_context=None, global_context=None, request=Request()):
        if local_context is None:
            local_context = {}
        if global_context is None:
            global_context = {}
        find_file_in_dirs(filename, self.directories)
        return Document(filename=filename, lookup=self,
                        local_context=local_context, global_context=global_context, request=request)

    def get_absolute_path(self, filename):
        find_file_in_dirs(filename, self.directories)
        return filename


def render(tempsource, local_context=None, global_context=None,
           path=None, get_params={}, post_params={}, base_indent=0, cid=None):
    if local_context is None:
        local_context = {}
    if tempsource.endswith('.zml'):
        return render_document(tempsource, local_context=local_context, global_context=global_context, path=path, cid=cid)
    else:
        return render_source(tempsource, local_context, global_context=global_context, base_indent=base_indent,
                             path=path, cid=cid)


def render_document(documentfile, local_context=None, global_context=None,
                    path=None, get_params={}, post_params={}, base_indent=0, cid=None):
    if local_context is None:
        local_context = {}
    lookup = DocumentLookup(['.'])
    request = Request(path=path)
    document = lookup.get_document(documentfile, local_context=local_context, global_context=global_context,
                                   request=request)
    out = document.render(base_indent=base_indent)
    return out


def render_source(source,
                local_context=None,
                global_context=None,
                indent_global='',
                source_indent_level=0,
                local_context_item='_root',
                path=None,
                get_params={},
                post_params={},
                base_indent=0, cid=None):
    if local_context is None:
        local_context = {}
    request = Request(path=path, get_params=get_params, post_params=post_params)
    document = Document(source=source, local_context=local_context, global_context=global_context, request=request,
                        cid=cid)
    root = document.source_to_tree(source)
    # we have to process the subtree to retrieve the routes
    # which are required by the components using the dispatcher
    root.process_subtree(Processor(lazy=True))
    out = root.render_subtree(Renderer())
    if document.inheriting_document:
        # here we set local_context of inheriting document to local_context of inherited document
        # todo: decide how we design the scopes of contexts
        if isinstance(document.inheriting_document, Resource):
            document.inheriting_document.load()
            inheriting_document = Document(source=document.inheriting_document.code, inherited_document=document,
                                           request=request)
            inheriting_document.dispatched_views = document.dispatched_views
            out = inheriting_document.render(local_context=document.local_context,
                                             base_indent=base_indent)
        else:
            inheriting_document = Document(document.inheriting_document, inherited_document=document, request=request)
            inheriting_document.dispatched_views = document.dispatched_views
            out = inheriting_document.render(local_context=document.local_context,
                                             base_indent=base_indent)
    return out


def load(tempsource, local_context=None, path=None,
         global_context=None, get_params={}, post_params={}, base_indent=0, lazy=False):
    if local_context is None:
        local_context = {}
    if tempsource.endswith('.zml'):
        return load_document(tempsource, local_context=local_context, global_context=global_context, path=path)
    else:
        return load_source(tempsource, local_context, global_context=global_context, base_indent=base_indent,
                           path=path, get_params=get_params, post_params=post_params, lazy=lazy)


def load_document(documentfile, local_context=None, path=None,
           global_context=None, get_params={}, post_params={}, base_indent=0):
    if local_context is None:
        local_context = {}
    request = Request(path=path, get_params=get_params, post_params=post_params)
    lookup = DocumentLookup(['.'])
    document = lookup.get_document(documentfile, local_context=local_context, global_context=global_context, request=request)
    document.import_document(request=request, base_indent=base_indent)
    root = document.source_to_tree(document.source)
    root.process_subtree(Processor(lazy=False))
    return document


def import_document(documentfile, local_context=None, path=None,
           global_context=None, get_params={}, post_params={}, base_indent=0):
    if local_context is None:
        local_context = {}
    request = Request(path=path, get_params=get_params, post_params=post_params)
    lookup = DocumentLookup(['.'])
    document = lookup.get_document(documentfile, local_context=local_context, global_context=global_context, request=request)
    document.import_document(request=request, base_indent=base_indent)
    return document


def import_file(documentfile, local_context=None, path=None,
           global_context=None, get_params={}, post_params={}, base_indent=0):
    if local_context is None:
        local_context = {}
    request = Request(path=path, get_params=get_params, post_params=post_params)
    lookup = DocumentLookup(['.'])
    document = lookup.get_document(documentfile, local_context=local_context, global_context=global_context, request=request)
    zmlsource = get_source_from_file(document.filename)
    return import_source(zmlsource)


def load_source(source,
                local_context=None, global_context=None,
                path=None, get_params={}, post_params={},
                indent_global='', source_indent_level=0, local_context_item='_root',
                base_indent=0, cid=None):
    # remoce leading and trailing whitespace, which is not syntax relevant
    source = source.strip()
    if local_context is None:
        local_context = {}
    request = Request(path=path, get_params=get_params, post_params=post_params)
    document = Document(source=source, local_context=local_context, global_context=global_context, request=request,
                        cid=cid)
    root = document.source_to_tree(source)
    root.process_subtree(Processor(lazy=False))
    if document.inheriting_document:
        # here we set local_context of inheriting document to local_context of inherited document
        # todo: decide how we design the scopes of contexts
        if isinstance(document.inheriting_document, Resource):
            document.inheriting_document.load()
            inheriting_document = Document(source=document.inheriting_document.code, inherited_document=document)
            inheriting_document.import_source()
        else:
            inheriting_document = Document(source=document.inheriting_document, inherited_document=document)
            inheriting_document.import_document()
    document.root = root
    return document


def import_source(source,
                  local_context=None, global_context=None,
                  path=None, get_params={}, post_params={},
                  indent_global='', source_indent_level=0, local_context_item='_root',
                  base_indent=0, cid=None, lazy=True):
    # remoce leading and trailing whitespace, which is not syntax relevant
    source = source.strip()
    if local_context is None:
        local_context = {}
    request = Request(path=path, get_params=get_params, post_params=post_params)
    document = Document(source=source, local_context=local_context, global_context=global_context, request=request,
                        cid=cid)
    root = document.source_to_tree(source)
    root.process_subtree(Processor(lazy=lazy))
    if document.inheriting_document:
        # here we set local_context of inheriting document to local_context of inherited document
        # todo: decide how we design the scopes of contexts
        if isinstance(document.inheriting_document, Resource):
            document.inheriting_document.load()
            inheriting_document = Document(source=document.inheriting_document.code, inherited_document=document)
            inheriting_document.import_source()
        else:
            inheriting_document = Document(source=document.inheriting_document, inherited_document=document)
            inheriting_document.import_document()
    document.root = root
    return document


def get_source_from_file(filename):
    lookup = DocumentLookup(['.'])
    document_path = lookup.get_absolute_path(filename)
    zmlsource = load_file(document_path)
    return zmlsource



from copy import deepcopy
from zml.semantic import *
from zml.context import RenderingContext
from zml.exceptions import IndentationException
from zml.model import Model
from zml.resource import Resource
import zml


num_spaces_per_indent = 2


def eval_context_item(res, node):
    value = None
    if 'resource' in res:
        document = zml.document.Document()
        resource = node.document.resources[res['resource']]
        resource.import_resource(document)
        value = document
        if 'data_node_descriptor' in res:
            value = value.get_var(res['data_node_descriptor'])
            if isinstance(value, Model):
                value = value
        elif 'model_descriptor' in res:
            value = value.models[res['model_descriptor']].fields
    elif 'context_item' in res and 'properties' not in res:
        value = node.get_var(res['context_item'])
        if isinstance(value, Model):
            value = value
    elif 'data_node_descriptor' in res:
        value = node.get_var(res['data_node_descriptor'])
        if isinstance(value, Model):
            value = value
        if isinstance(value, Field):
            value = value
    elif 'model_descriptor' in res:
        value = node.document.models[res['model_descriptor']].fields
    if 'properties' in res:
        # in some cases there is a context_item defined in res
        # todo specify different cases
        if 'context_item' in res:
            value = node.get_var(res['context_item'])
        if not value:
            raise Exception('There is no context item named {}'.format(res['context_item']))
        if not isinstance(value, dict) and not isinstance(value, Field) and not isinstance(value, Model):
            try:
                # check if local context exists (otherwise use the main object instead)
                value = value.local_context
            except:
                pass
        properties = res['properties']
        value = get_combined_properties(value, properties)
    return value


def get_property(obj, property_descriptor):
    # check if property in local_context, othterwise use main object
    try:
        value = obj[property_descriptor]
    except:
        value = getattr(obj, property_descriptor)
    return value


def get_meta_data(obj, property_descriptor):
    if property_descriptor == 'descriptor':
        value = obj.descriptor
    elif property_descriptor == 'label':
        value = obj.label
    elif property_descriptor == 'type':
        value = obj.type
    return value


def get_combined_properties(value, properties):
    for property_item in properties:
        # currently we only handle dict properties. future implementations will handle
        # list item accessors with dot-number syntax (.0 for first element, .1 for second etc.)
        property_descriptor = property_item[1][0]
        if property_item[0] == 'property':
            value = get_property(value, property_descriptor)
        elif property_item[0] == 'meta_data':
            value = get_meta_data(value, property_descriptor)
    return value


def eval_model(res, node):
    if 'model_descriptor' in res:
        return node.document.models[res['model_descriptor']]


def eval_translation(res, node):
    translation_descriptor = res['translation_descriptor']
    if 'properties' in res:
        value = node.document.get_translation(translation_descriptor, node.document.language)
        for property_item in res['properties']:
            property_descriptor = property_item[1][0]
            value = value[property_descriptor]
    else:
        value = node.document.get_translation(translation_descriptor, node.document.language)
    return value


def render_translation(res, node):
    return str(eval_translation(res, node))


class NodeRenderingContext(RenderingContext):
    pass


class Path(object):

    def __init__(self, document):
        self.document = document

    def execute(self, context, *args, **kwargs):
        if 'action' in context['context']:
            action = context['context']['action']
        else:
            return ''
        if 'router' in context['context']:
            router = self.document.router[context['context']['router']]
        else:
            router = self.document.default_router
        if router and action in router:
            route = router[action]
            url = ''
            for item in route:
                if 'path_segment' in item:
                    url += '/' + item[1][0]

                elif 'path_variable' in item:
                    path_variable = item[1][0]
                    if path_variable in context['context']:
                        url += '/' + context['context'][path_variable]
            return url


framework_components = {
    'path': Path,
    'sin': math.sin,
    'pi': math.pi
}


class TreeNode:

    def __init__(self, line='', line_number=1, is_root=False, is_ancestor=False, ancestor=None, base_indent=0):
        self.line = line
        self.base_indent = base_indent
        self.children = []
        self.body = None
        self.value = None
        self.link = None
        self.is_root = is_root
        self.is_ancestor = is_ancestor
        if ancestor is None:
            # explicitly state logic here to be clear that in both cases ancestor is set to self
            if self.is_ancestor:
                self.ancestor = self
            else:
                self.ancestor = self
        else:
            self.ancestor = ancestor
        indentation = len(line) - len(line.lstrip())
        if indentation % num_spaces_per_indent != 0:
            raise IndentationException('Wrong indentation in line: {}'.format(line_number))
        self.level = int(indentation / num_spaces_per_indent)
        self.render_level = 0
        self.base_render_level = 0

    def __repr__(self):
        # return '{}\n{}'.format(self.line, json.dumps(self.local_context, indent=2))
        return "{}({})\n".format(self.__class__.__name__, self.line)

    def add_children(self, nodes):
        if nodes:
            childlevel = nodes[0].level
            while nodes:
                node = nodes.pop(0)
                node.ancestor = self
                if node.level == childlevel:
                    if self.is_ancestor:
                        node.ancestor = self
                        node.local_context = {}
                    # add node as a child
                    self.children.append(node)
                elif node.level > childlevel:
                    # add nodes as grandchildren of the last child
                    # if self.children[-1].is_ancestor:
                    #     node.ancestor = self.children[-1]
                    #     node.local_context = self.children[-1].local_context
                    nodes.insert(0, node)
                    self.children[-1].add_children(nodes)
                elif node.level <= self.level:
                    # this node is a sibling, no more children
                    nodes.insert(0, node)
                    return

    def is_data(self):
        if self.is_root:
            return False
        else:
            return self.parent.is_data()


class Egg(TreeNode):

    expression = EGG

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        res = self.expression.parseString(self.line)
        if 'key' in res:
            self.body = res['key']
        if 'value' in res:
            self.link = res['value']
        # print('***')
        # print(self.line)
        # print(self.value)


class Node(TreeNode, NodeRenderingContext):

    expression = ELEMENT

    def __init__(self, line, document=None, expression=None, global_context=None, local_context=None,
                 meta_data={}, descriptor=None, base_indent=0, renderer=None, *args, **kwargs):
        super().__init__(line, *args, **kwargs)
        if local_context is None:
            local_context = {}
        if global_context is None:
            global_context = {}
        self.line = line
        self.data = dict()
        self.descriptor = descriptor
        self.local_context = local_context
        self.meta_data = meta_data
        self.global_context = global_context
        self.renderer = renderer
        self.minimise = False
        self.leading = ''
        self.trailing = ''
        self.is_component = False
        self.is_code = False
        self.has_list_items = False
        self.line = line.strip()
        self.document = document
        self.out = ()
        if self.is_root:
            self.level = -1

    def get_path(self):
        node = self
        path = ''
        while node != self.document.root:
            segment = ''
            if node.glyph:
                segment += node.glyph
            segment += node.descriptor
            path = segment + path
            if node.parent:
                node = node.parent
            else:
                return path
        return path

    def eval_segment(self, segment):
        if 'data_node_descriptor' in segment[1]:
            return self.data[segment[1]['data_node_descriptor']]

    def get_address(self):
        if self.document.cid:
            return self.document.cid + self.get_path()
        else:
            return self.get_path()

    def process(self, lazy=False):
        if not lazy and self.is_root is False:
            self.expression.setParseAction(self.parse)
            res = self.expression.parseString(self.line)
            self.tokens = self.expression.parseString(self.line)
            self.render_parts(res)

    def render(self):
        self.tokens = self.expression.parseString(self.line)
        if 'inline_content' in self.tokens:
            value = self.render_inline_content(self.tokens)
            self.set_var('_value', value)
        if self.is_root is False:
            self.expression.setParseAction(self.parse)
            res = self.expression.parseString(self.line)
            self.render_parts(res)
        else:
            return ''

    def get_combined_properties(self, properties):
        return get_combined_properties(self, properties)

    def get_meta_data(self, property_descriptor):
        return get_meta_data(self, property_descriptor)

    def eval_context_item(self, res):
        return eval_context_item(res, self)

    def is_list(self):
        return self.has_list_items

    def parse(self, s, l, t):
        return t

    def render_parts(self, res):
        self.leading = self.render_leading(res)
        self.trailing = self.render_trailing(res)

    def render_indent(self):
        render_level = (self.ancestor.base_render_level + self.render_level)
        indent = ' ' * self.base_indent + ' ' * render_level * 2
        return indent

    def render_start_tag(self, res):
        descriptor = self.render_descriptor(res)
        attributes = self.render_attributes(res)
        id_classes = self.render_id_classes(res)
        out = '<{}{}{}>'.format(descriptor, id_classes, attributes)
        return out

    def render_inline_start_tag(self, res):
        descriptor = self.render_descriptor(res)
        attributes = self.render_attributes(res)
        id_classes = self.render_id_classes(res)
        value = ''
        if 'inline_content' in res:
            value = self.render_inline_content(res)
        elif 'expression' in res:
            value = self.eval_expression(res)
        out = '<{}{}{}>{}'.format(descriptor, id_classes, attributes, value)
        return out

    def render_end_tag(self, res):
        descriptor = self.render_descriptor(res)
        out = ''
        if descriptor in void_elements:
            out += ''
        else:
            if not self.minimise and descriptor not in inline_elements:
                render_level = (self.ancestor.base_render_level + self.render_level)
                out += ' ' * self.base_indent + ' ' * render_level * 2
            out += '</{}>'.format(descriptor)
            if not self.minimise:
                out += '\n'
        return out

    def render_inline_end_tag(self, res):
        descriptor = self.render_descriptor(res)
        out = '</{}>'.format(descriptor)
        return out

    def render_inline_semantics(self, res):
        out = ''
        out += self.render_inline_start_tag(res)
        out += res['inline_semantic_content_words']
        out += self.render_inline_end_tag(res)
        if 'trailing_space' in res:
            out += ' '
        return out

    def render_leading(self, res):
        descriptor = self.render_descriptor(res)
        start_tag = self.render_start_tag(res)
        value = ''
        if 'inline_content' in res:
            value = self.render_inline_content(res)
        elif 'expression' in res:
            value = self.eval_expression(res)
        leading = ''
        if not self.minimise:
            leading += self.render_indent()
        leading += start_tag + value
        if not self.minimise and descriptor not in inline_elements:
            leading += '\n'
        return leading

    def render_trailing(self, res):
        return self.render_end_tag(res)

    def render_descriptor(self, res):
        if 'descriptor' in res:
            return res['descriptor']
        else:
            return ''

    def eval_arith_expression(self, res):
        expr = ''.join(res['arith_expression'])
        del exprStack[:]
        ARITH_EXPRESSION.parseString(expr, parseAll=True)
        return evaluateStack(exprStack)

    def eval_boolean(self, res):
        if res['boolean'] == 'True':
            return True
        elif res['boolean'] == 'False':
            return False

    def render_boolean(self, res):
        return str(self.eval_boolean(res))

    def eval_literal(self, res):
        # note that this works by using addParseAction(removeQuotes) in LITERAL
        value = res['literal']
        return value

    def render_literal(self, res):
        return str(self.eval_literal(res))

    def eval_expression(self, res):
        if 'arith_expression' in res:
            return self.eval_arith_expression(res)
        elif 'boolean' in res:
            return self.eval_boolean(res)
        elif 'literal' in res:
            return self.eval_literal(res)
        elif 'framework_component' in res:
            return self.eval_framework_component(res)
        elif 'component_descriptor' in res:
            return self.eval_component(res)
        elif 'value_accessor' in res:
            # might lead to endless loop ?
            return self.eval_value_accessor(res)
        elif 'meta_data' in res or 'resource' in res or 'context_item' in res or 'context_item_with_property' in res:
            return self.eval_context_item(res)
        elif 'translation_accessor' in res:
            return eval_translation(res, self)

    def eval_attribute_value(self, attribute):
        if 'moustache' in attribute:
            attribute_value = self.render_moustache(attribute)
        elif 'quoted_string' in attribute:
            attribute_value = attribute['quoted_string']
        elif 'number' in attribute:
            attribute_value = attribute['number']
        elif 'literal' in attribute:
            attribute_value = attribute['literal']
        elif 'context_item' in attribute:
            return self.eval_context_item(attribute)
        elif 'model_descriptor' in attribute:
            return eval_model(attribute, self)
        return attribute_value

    def render_attribute(self, attribute):
        out = ' '
        out += attribute['attribute_key']
        if 'attribute_value' in attribute:
            out += '="{}"'.format(self.eval_attribute_value(attribute))
        return out

    def render_attributes(self, res):
        if 'attributes' in res:
            return ' '.join(
                [''.join(
                    [self.render_attribute(attribute) for attribute in res['attributes']]
                )]
            )
        else:
            return ''

    def render_id_classes(self, res):
        out = ''
        if 'id_classes' in res:
            if 'uid' in res:
                out += ' id="{}"'.format(res['uid'][0])
            if 'classes' in res:
                out += ' class="{}"'.format(' '.join(res['classes']))
        return out

    def render_inline_content(self, res):
        # todo: add inline slot rendering
        out = ''
        if 'inline_content' in res:
            inline_content_items = res['inline_content']
            for item in inline_content_items:
                itemtype = item[0]
                if itemtype == 'inline_content_words':
                    out += self.render_words(item[1][0])
                elif itemtype == 'inline_content_newlines':
                    out += ''.join(item[1])
                if itemtype == 'inline_semantics':
                    out += self.render_inline_semantics(item[1])
                if itemtype == 'moustache':
                    out += self.render_moustache(item[1])
#                if itemtype == 'translation_accessor':
#                    out += render_translation(item[1], self)
        return out

    def render_words(self, res):
        out = res
        return out

    def eval_value_accessor(self, res):
        return self.get_var('_value')

    def render_value_accessor(self, res):
        return self.eval_value_accessor(res)

    def _render_component(self, component_descriptor):
        namespace = self.document.namespace
        component = None
        # rendered_view_source = local_indent
        rendered_view_source = ''
        if component_descriptor in self.document.namespaces[namespace]:
            component = self.document.namespaces[namespace][component_descriptor]
        elif component_descriptor in self.inherited_document.namespaces[namespace]:
            component = self.inherited_document.namespaces[namespace][component_descriptor]
        if component:
            component_node = deepcopy(component)
            component_node.parent = self
            # component_node.ancestor = self.ancestor
            component_node.base_render_level = self.render_level
            if not self.parent.is_component and not self.parent.is_code:
                component_node.base_render_level += 1

            component_node.render_level = component_node.base_render_level
            # local_indent = ' ' * self.render_level * 2
            self.renderer.base_render_level = self.render_level
            rendered_component = component_node.render_subtree(self.renderer)
            rendered_view_source += rendered_component
        # strip last newline, as it would double with the inline content's trailing newline
        rendered_view_source = rendered_view_source.rstrip()
        return rendered_view_source

    def render_component(self, res):
        # prepend local indent on each line
        component_descriptor = res['component_descriptor']
        # prepend newline, because view will be included in inline content,
        # which is missing a leading newline
        return '\n' + self._render_component(component_descriptor)

    def eval_framework_component(self, res):
        framework_component = res['framework_component']
        if 'moustache_attributes' in res:
            params = dict()
            for attribute in res['moustache_attributes']:
                key = attribute['moustache_attribute_key']
                value = attribute['moustache_attribute_value']
                if value[0] == '_context':
                    value = self.local_context
                params[key] = value
            return framework_components[framework_component](self.document).execute(params)

    def render_framework_component(self, res):
        return self.eval_framework_component(res)

    def render_context_item(self, res):
        return str(self.eval_context_item(res))

    def render_moustache(self, res):
        out = ''
        if 'framework_component' in res:
            out += self.render_framework_component(res)
        elif 'component_descriptor' in res:
            out += self.render_component(res)
        elif 'value_accessor' in res:
            out += self.render_value_accessor(res)
        elif 'context_item' in res:
            out += self.render_context_item(res)
        elif 'context_item_with_property' in res:
            out = self.render_context_item(res)
        elif 'translation_accessor' in res:
            out += render_translation(res, self)
        return out

    def process_subtree(self, processor=None):
        processor.process(self)

    def render_subtree(self, renderer=None):
        out = renderer.render(self)
        return out

    def get_children_type(self):
        return None


class EmptyNode(Node):

    def render_leading(self, res):
        descriptor = self.render_descriptor(res)
        start_tag = ''
        inline_content = self.render_inline_content(res)
        leading = ''
        if not self.minimise:
            leading += self.render_indent()
        leading += start_tag + inline_content
        if not self.minimise and descriptor not in inline_elements:
            leading += '\n'
        return leading


class ComponentNode(Node):

    expression = LINE
    glyph = '*'

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.is_component = True
        self.caller_inline_content = None
        self.caller_children = None

    def process(self, lazy=False):
        self.expression.setParseAction(self.parse)
        self.tokens = self.expression.parseString(self.line)
        descriptor = self.tokens['descriptor']
        # self.local_context[descriptor] = self.childsource
        if self.document.namespace not in self.document.namespaces:
            self.document.namespaces[self.document.namespace] = dict()
        self.document.namespaces[self.document.namespace][descriptor] = self

    def render(self):
        pass


class ComponentCallNode(Node):

    expression = COMPONENT_CALL
    glyph = '*'

    def process(self, lazy=False):
        if not lazy:
            self.expression.setParseAction(self.parse)
            self.tokens = self.expression.parseString(self.line)

    def render(self):
        self.process()
        namespace = self.tokens['namespace']
        descriptor = self.tokens['component_descriptor']
        component_node = deepcopy(self.document.namespaces[namespace][descriptor])
        component_node.parent = self.parent
        component_node.ancestor = self.ancestor
        # the component call inherits the total render_level of parent and the parent's ancestor to the component
        component_node.base_render_level = self.parent.render_level + self.parent.ancestor.base_render_level
        if 'inline_content' in self.tokens:
            component_node.caller_inline_content = self.tokens['inline_content']
        component_node.caller_children = self.children
        self.renderer.base_render_level = self.render_level
        if not self.parent.is_component and not self.parent.is_code:
            component_node.base_render_level += 1
        if 'attributes' in self.tokens:
            for attribute in self.tokens['attributes']:
                attribute_key = attribute['attribute_key']
                attribute_value = self.eval_attribute_value(attribute)
                component_node.set_var(attribute_key, attribute_value)
        if 'inline_content' in self.tokens:
            value = self.render_inline_content(self.tokens)
            component_node.set_var('_value', value)
        children = []
        for child in self.children:
            children.append(child.render_subtree(self.renderer))
        component_node.set_var('_children', children)
        out = ''
        out += component_node.render_subtree(self.renderer)
        self.leading = out
        return out


class ListItemNode(Node):

    expression = LIST_ITEM

    def process(self, lazy=False):
        if not lazy:
            self.expression.setParseAction(self.parse)
            self.tokens = self.expression.parseString(self.line)
            self.parent.has_list_items = True
            # if self.parent.descriptor not in self.local_context:
            #    self.set_var(self.parent.descriptor, list())
            if not isinstance(self.parent.value, list):
                self.parent.value = list()
            if isinstance(self.parent, DataNode):
                # self.document.local_context[self.parent.descriptor] = self.parent.value
                self.local_context[self.parent.descriptor] = self.parent.value
            if 'inline_content' in self.tokens:
                self.value = self.render_inline_content(self.tokens)
            elif 'expression' in self.tokens:
                self.value = self.eval_expression(res)
            else:
                self.value = dict()
            self.parent.value.append(self.value)

    def render(self):
        self.process()


class TranslationNode(Node):

    expression = TRANSLATION
    glyph = '!'

    def process(self, lazy=False):
        self.value = dict()
        self.document.translations[self.descriptor] = self.value

    def render(self):
        self.process()

    def get_children_type(self):
        return AssignmentNode


class AssignmentNode(Node):

    expression = ASSIGNMENT

    def process(self, lazy=False):
        res = self.expression.parseString(self.line)
        descriptor = res['descriptor']
        value = ''
        if 'inline_content' in res:
            value = self.render_inline_content(res)
        elif 'expression' in res:
            value = self.eval_expression(res)
        # self.parent.set_var(descriptor, value)
        if isinstance(self.parent, AssignmentNode) or isinstance(self.parent, ListItemNode) or isinstance(self.parent,
                                                                                                          TranslationNode):
            if self.parent.value is None:
                self.parent.value = dict()
            if value:
                self.parent.value[descriptor] = value
            else:
                self.value = dict()
                self.parent.value[descriptor] = self.value
        elif isinstance(self.parent, DataNode):
            if self.local_context[self.parent.descriptor] in [None, '']:
                self.local_context[self.parent.descriptor] = dict()
            if value:
                self.local_context[self.parent.descriptor][descriptor] = value
            else:
                self.value = dict()
                self.local_context[self.parent.descriptor][descriptor] = self.value

    def render(self):
        self.process()


class FieldPropertyNode(Node):

    expression = ASSIGNMENT

    def process(self, lazy=True):
        res = self.expression.parseString(self.line)
        descriptor = res['descriptor']
        glyph = res['glyph']
        if 'inline_content' in res:
            value = self.render_inline_content(res)
        elif 'expression' in res:
            value = self.eval_expression(res)
        if glyph == '&':
            # define meta data
            if descriptor == 'type':
                self.parent.value.type = value
            if descriptor == 'label':
                self.parent.value.label = value

    def render(self):
        self.process()


class Field:

    def __init__(self, descriptor):
        self.descriptor = descriptor
        self.type = None
        self.label = ''


class FieldNode(Node):

    expression = FIELD

    def process(self, lazy=False):
        res = self.expression.parseString(self.line)
        descriptor = res['field_descriptor']
        self.value = Field(descriptor)
        self.parent.value.fields[descriptor] = self.value

    def render(self):
        self.process()

    def get_children_type(self):
        return FieldPropertyNode


class ModelNode(Node):

    expression = MODEL
    glyph = '+'

    def process(self, lazy=False):
        res = self.expression.parseString(self.line)
        descriptor = res['model_descriptor']
        self.value = Model(descriptor)
        self.document.models[self.descriptor] = self.value

    def render(self):
        self.process()

    def get_children_type(self):
        return FieldNode


class ResourceNode(Node):

    expression = RESOURCE
    glyph = '@'

#    def __init__(self):
#        super(ResourceNode, self).__init__()
#        if self.line:
#            res = self.expression.parseString(self.line)
#            self.address = res['source']
#            self.resource = Resource(address, lazy=True)

    def process(self, lazy=False):
        res = self.expression.parseString(self.line)
        address = res['source']
        self.resource = Resource(address)
        self.document.resources[self.descriptor] = self.resource

    def render(self):
        self.process()


class RouteNode(Node):

    expression = ROUTE

    def process(self, lazy=False):
        self.expression.setParseAction(self.parse)
        res = self.expression.parseString(self.line)
        router_descriptor = self.parent.descriptor
        route_descriptor = res['descriptor']
        # currently also nodes have a global_context and a local_context
        # todo design decision for rendering contexts (see module's render_source function)
        router = self.document.router
        if router_descriptor not in router:
            router[router_descriptor] = dict()

            router[router_descriptor][route_descriptor] = res.get('path_items')
            self.document.default_router = router[router_descriptor]
        else:
            router[router_descriptor][route_descriptor] = res.get('path_items')

    def render(self):
        self.process()


class RouterNode(Node):

    expression = LINE
    glyph = '~'

    def process(self, lazy=False):
        pass

    def get_children_type(self):
        return RouteNode

    def render(self):
        self.process()
        # we have to overwrite render to prevent rendering as Node
        pass


class DataNode(Node):

    expression = DATA_NODE
    glyph = '#'

    def is_data(self):
        return True

    def process(self, lazy=False):
        res = self.expression.parseString(self.line)
        descriptor = res['data_descriptor']
        value = ''
        if 'inline_content' in res:
            value = self.render_inline_content(res)
        elif 'expression' in res:
            value = self.eval_expression(res)
        descriptor = res['data_descriptor']
        self.local_context[descriptor] = value

    def render(self):
        self.process()

    # def get(self, path):
    #    res = COMBINED_ACCESSOR.parseString(path)
    #    return res


class MetaNode(Node):

    expression = LINE
    glyph = '&'

    def is_meta(self):
        return True

    def process(self, lazy=False):
        res = self.expression.parseString(self.line)
        descriptor = res['descriptor']
        self.meta_data[descriptor] = self.value

    def render(self):
        self.process()


class SlotNode(Node):

    expression = SLOT_NODE
    glyph = '|'

    def __init__(self, *args, **kwargs):
        self.descriptor = kwargs['descriptor']
        super().__init__(*args, **kwargs)

    def render(self):
        if self.descriptor in self.document.dispatched_views:
            view_name = self.document.dispatched_views[self.descriptor]
            namespace = self.document.namespaces[self.document.namespace]
            if view_name in namespace:
                rendered_component = self._render_component(view_name)
                self.leading = rendered_component + '\n'
                return ''  # rendered_component
        else:
            return ''


class CodeNode(Node):

    expression = CODE
    glyph = '%'

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.is_code = True

    def is_instruction(self):
        if self.descriptor in ['inherit', 'import', 'include', 'namespace']:
            return True
        else:
            return False

    def process_subtree(self, processor):
        self.expression.setParseAction(self.parse)
        self.tokens = self.expression.parseString(self.line)
        if 'for_loop' in self.tokens:
            processor.subtree_handler_for_loop(self)
        elif 'if_statement' in self.tokens:
            processor.subtree_handler_if_statement(self)
        if self.is_instruction():
            processor.process(self)

    def render_subtree(self, renderer):
        self.expression.setParseAction(self.parse)
        self.tokens = self.expression.parseString(self.line)
        out = ''
        if 'for_loop' in self.tokens:
            out += renderer.subtree_handler_for_loop(self)
        elif 'if_statement' in self.tokens:
            out += renderer.subtree_handler_if_statement(self)
        if self.is_instruction():
            out += renderer.render(self)
        return out

    def process_inherit(self, res):
        if 'documentfile' in res:
            documentfile = res['documentfile']
            if not documentfile.endswith('.zml'):
                documentfile += '.zml'
            self.document.inheriting_document = documentfile
        elif 'resource' in res:
            resource_name = res['resource']
            if resource_name in self.document.resources:
                self.document.inheriting_document = self.document.resources[resource_name]

    def process_import(self, res):
        if 'documentfile' in res:
            documentfile = res['documentfile']
            if not documentfile.endswith('.zml'):
                documentfile += '.zml'
            self.document.import_document(documentfile)
        elif 'resource' in res:
            resource_descriptor = res['resource']
            self.document.resources[resource_descriptor].import_resource(self.document)
        # return root

    def process_include(self, res):
        out = render(filename,
                     local_context=self.document.local_context,
                     base_indent=base_indent)
        return out

    def process_namespace(self, res):
        namespace = res['namespace']
        self.document.namespace = namespace

    def process(self, lazy=False):
        self.expression.setParseAction(self.parse)
        res = self.expression.parseString(self.line)
        if 'instruction' in res:
            instruction = res['instruction']
            if instruction == 'inherit':
                self.process_inherit(res)
            elif instruction == 'import':
                self.process_import(res)
            elif instruction == 'include':
                self.process_include(res)
            elif instruction == 'namespace':
                self.process_namespace(res)

    def render(self, lazy=False):
        self.process(lazy)
        return ''



#!/usr/bin/env python
from setuptools import setup
setup(
    name='zml',
    packages=['zml'],
    version='0.10.15',
    description='zero markup language',
    author='Christof Hagedorn',
    author_email='zeromage@neo.codes',
    url='http://www.zml.org/',
    download_url='https://pypi.python.org/pypi/zml',
    keywords=['zml', 'zero', 'markup', 'language', 'template', 'templating'],
    # install_requires=open("requirements/base.in").readlines(),
    classifiers=[
        'Programming Language :: Python',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.2',
        'Development Status :: 3 - Alpha',
        'Environment :: Web Environment',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: BSD License',
        'Operating System :: OS Independent',
        'Topic :: Internet',
        'Topic :: Internet :: WWW/HTTP',
        'Topic :: Internet :: WWW/HTTP :: Dynamic Content',
        'Topic :: Internet :: WWW/HTTP :: WSGI',
        'Topic :: Software Development :: Libraries :: Python Modules',
    ],
    long_description='''\
zml - zero markup language
-------------------------------------

Features
 - zero markup templates
 - clean syntax
 - extensible
 - components
 - namespaces
 - lean code
 - rendering of rest data
 - transclusions
 - integration of external defined objects (IPFS)
 - sustainable linked data by using immutable objects (decentral storage)

This version requires Python 3 or later.
'''
)



import requests
import ipfshttpclient
from functools import lru_cache
import functools
from datetime import datetime, timedelta
import magic
from zml.crypt import *
import json


def timed_cache(**timedelta_kwargs):

    def _wrapper(f):
        update_delta = timedelta(**timedelta_kwargs)
        next_update = datetime.utcnow() - update_delta
        # Apply @lru_cache to f with no cache size limit
        f = functools.lru_cache(None)(f)

        @functools.wraps(f)
        def _wrapped(*args, **kwargs):
            nonlocal next_update
            now = datetime.utcnow()
            if now >= next_update:
                f.cache_clear()
                next_update = now + update_delta
            return f(*args, **kwargs)
        return _wrapped
    return _wrapper


class Backend:

    def load(self):
        raise NotImplementedError("Method load not implemented.")

    def save(self):
        raise NotImplementedError("Method save not implemented.")


class FileBackend(Backend):

    def load(self, filename):
        with open(filename) as f:
            raw = f.read()
        mime_type = magic.from_buffer(raw, mime=True)
        return raw, mime_type

    def save(self, filename, code):
        with open(filename, 'w', encoding='utf-8') as f:
            f.write(code)


class IpfsBackend(Backend):

    def __init__(self):
        self.client = ipfshttpclient.connect('/ip4/127.0.0.1/tcp/5001/http')

    @lru_cache()
    def get_peer_id(self):
        return self.client.key.list()['Keys'][0]['Id']

    @lru_cache()
    def get_peer_cid(self):
        peer_id = self.get_peer_id()
        profile_cid = self.client.name.resolve(peer_id)['Path'].split('/')[-1]
        return profile_cid

    @lru_cache()
    def get_profile(self):
        profile_cid = self.get_peer_cid()
        return self.client.cat(profile_cid).decode('utf-8')

    def cache_clear(self):
        self.get_peer_id.cache_clear()
        self.get_profile.cache_clear()

    def add_str(self, code):
        return self.client.add_str(code)

    def add_bytes(self, code):
        return self.client.add_bytes(code)

    def load(self, address):
        if address.startswith('/ipns/'):
            address = self.client.resolve(address)['Path']
        raw = self.client.cat(address)
        mime_type = magic.from_buffer(raw, mime=True)
        if mime_type in ['image/jpeg', 'image/png', 'image/gif']:
            res = raw
        else:
            try:
                res = raw.decode('utf-8')
                mime_type = 'text/zml'
            except:
                return raw, 'crypt/zml'
        return res, mime_type

    def encrypt(self, raw, key):
        return encrypt(raw, key)

    def decrypt(self, raw, key):
        return decrypt(raw, key)

    def save(self, code):
        if isinstance(code, str):
            return self.add_str(code)
        elif isinstance(code, bytes):
            return self.add_bytes(code)
        else:
            raise "Content must be string or bytes"

    def publish(self, ipfs_path, key=None):
        return self.client.name.publish(ipfs_path, key=key)

    def generate_key(self, name):
        try:
            status = self.client.key.gen(name, type='ed25519')
        except:
            return self.get_key(name)
        return status

    def get_key(self, name):
        status = self.client.key.list()
        if 'Keys' in status:
            for item in status['Keys']:
                if 'Name' in item and item['Name'] == name:
                    return item

    def resolve(self, name, stream=False):
        if stream:
            r = requests.get('http://localhost:5001/api/v0/name/resolve?arg={}&stream=1'.format(name), stream=True)
            for line in r.iter_lines():
                if line:
                    res = json.loads(line)
                    if 'Path' in res:
                        return res['Path']
        else:
            return self.client.name.resolve(name)['Path']


class RestBackend(Backend):

    def load(self, address):
        url = address
        if not url.startswith('https'):
            url = 'https://' + url
            response = requests.get(url)
        root = response.json()
        mime_type = 'text/html'
        return root, mime_type

    def save(self, code):
        pass



class DocumentNotDefinedException(Exception):
    message = 'Call to get_absolute_path without document parameter and no document set'


class FileNotLoadedException(Exception):
    message = 'Could not load file'


class IndentationException(Exception):
    message = 'Indentation not a multiple of two'


class VariableNotDefinedException(Exception):
    message = 'Variable not defined'


class TranslationNotDefinedException(Exception):
    message = 'Translation not defined'


class NotFoundException(Exception):
    message = 'Route can not be resolved'
